#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" By the proxy of launching product fetcher at the workflow container command
line, ensure we haven't broken the download capability """

import json
import logging
import os
import subprocess
import sys
from pathlib import Path
from typing import List
from unittest.mock import patch

import pytest
from productfetcher.interfaces import LocationReport
from productfetcher.locations import FileLocator

# pylint: disable=E0401, E0402, R0914


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

REALLY_FETCH = False

ASDM_BIN_SUBDIR = "ASDMBinary"

# TODO: test optional command-line parameters?


@pytest.mark.skipif(not REALLY_FETCH, reason="actually fetches")
def test_single_bdf_download(tmpdir):
    """
    Really, truly stream all 7 freakin' MB of a genuine ALMA BDF

    :return:
    """
    size = 7093883

    # create a locations file containing a single, relatively small BDF
    location_info = {
        "files": [
            {
                "ngas_file_id": "A002/Xe79954/X9711",
                "subdirectory": "uid___A002_Xe79954_X8f3f",
                "relative_path": "ASDMBinary/uid___A002_Xe79954_X9711",
                "checksum": "-1713733331",
                "checksum_type": "ngamsGenCrc32",
                "version": 1,
                "size": 7093883,
                "server": {
                    "server": "nangas14.cv.nrao.edu:7772",
                    "location": "NAASC",
                    "cluster": "NAASC",
                },
            },
        ],
        "aggregate_size": size,
    }
    json_str = json.dumps(location_info)

    os.chdir(Path(tmpdir))
    json_file = Path("alma_single_bdf_locations.json")
    with open(json_file, "w") as out:
        out.write(json_str)

    cmd_line_args = ["productfetcher", "--location-file", str(json_file)]
    # this is what we'd do IRL, but we don't want to hit NGAS in tests
    with subprocess.Popen(cmd_line_args) as proc:
        logger.debug(proc)

    dl_files = [file for file in Path(tmpdir).rglob("*") if file.is_file() and not file.name.endswith(".json")]
    assert len(dl_files) > 0
    bdf = dl_files[0]
    assert bdf.stat().st_size == size


def test_single_sdm_download(tmpdir):
    """
    Pretend to fetch a VLA SDM from a locations file.

    :return:
    """

    size = 380
    filename = "CalData.xml"

    # create a locations file containing a single VLA SDM
    location_info = {
        "files": [
            {
                "ngas_file_id": "uid___evla_sdm_X1391196414980.sdm",
                "subdirectory": "13B-014.sb28677984.eb28693176.56688.810349050924",
                "relative_path": filename,
                "checksum": "-1661730835",
                "checksum_type": "ngamsGenCrc32",
                "version": 1,
                "size": size,
                "server": {
                    "server": "nmngas04.aoc.nrao.edu:7777",
                    "location": "DSOC",
                    "cluster": "DSOC",
                },
            },
        ],
        "aggregate_size": size,
    }
    json_str = json.dumps(location_info)

    os.chdir(Path(tmpdir))
    json_file = Path("vla_single_sdm_locations.json")
    with open(json_file, "w") as out:
        out.write(json_str)

    cmd_line_args = ["productfetcher", "--location-file", str(json_file)]

    with patch("subprocess.Popen") as mock_proc:
        with patch(
            "productfetcher.fetchers.FileFetcher.do_fetch",
            return_value=mock_single_fetch(tmpdir, filename, size),
        ) as mock_fetch:
            execute_command(cmd_line_args)

            mock_proc.assert_called_once()
            mock_fetch.assert_not_called()

    dl_files = [file for file in Path(tmpdir).rglob("*") if file.name.endswith(".xml")]
    assert len(dl_files) == 1
    sdm = dl_files[0]
    assert sdm.stat().st_size == size


def test_alma_single_sdm_mocked(tmpdir):
    """
    Pretend to download an ALMA SDM from a locations file at the container
    command line.

    :return:
    """
    filename = "AnnotationTable.xml"
    size = 4215

    # create a locations file containing a single ALMA SDM
    location_info = {
        "files": [
            {
                "archive_uid": "uid://A002/Xe7b106/X1",
                "table": "XML_ANNOTATIONTABLE_ENTITIES",
                "relative_path": filename,
                "subdirectory": "uid___A002_Xe79954_X8f3f",
                "size": size,
            },
        ],
        "aggregate_size": size,
    }
    json_str = json.dumps(location_info)
    json_file = Path(tmpdir) / "alma_single_sdm_locations.json"
    with open(json_file, "w") as out:
        out.write(json_str)

    os.environ["CAPO_PATH"] = "/home/ssa/capo"
    os.environ["CAPO_PROFILE"] = "dsoc-dev"
    cmd_line_args = ["productfetcher", "--location-file", str(json_file)]

    with patch("subprocess.Popen") as mock_proc:
        with patch(
            "productfetcher.fetchers.FileFetcher.do_fetch",
            return_value=mock_single_fetch(tmpdir, filename, size),
        ) as mock_fetch:
            execute_command(cmd_line_args)

            mock_proc.assert_called_once()
            mock_fetch.assert_not_called()

    dl_files = [file for file in Path(tmpdir).rglob("*") if file.name.endswith(".xml")]
    assert len(dl_files) == 1
    sdm = dl_files[0]
    assert sdm.stat().st_size == size


@pytest.mark.skipif(not REALLY_FETCH, reason="actually fetches")
def test_alma_single_sdm_dl_4_realz(tmpdir):
    """
    Really retrieve a real SDM table from the real ALMA metadata DB that we
    really love, really

    :return:
    """

    size = 4215

    # create a locations file containing a single ALMA SDM
    location_info = {
        "files": [
            {
                "archive_uid": "uid://A002/Xe7b106/X1",
                "table": "XML_ANNOTATIONTABLE_ENTITIES",
                "relative_path": "AnnotationTable.xml",
                "subdirectory": "uid___A002_Xe79954_X8f3f",
                "size": size,
            },
        ],
        "aggregate_size": size,
    }
    json_str = json.dumps(location_info)

    output_dir = Path(tmpdir)
    json_file = output_dir / "alma_single_sdm_locations.json"
    with open(json_file, "w") as out:
        out.write(json_str)

    os.environ["CAPO_PATH"] = "/home/ssa/capo"
    os.environ["CAPO_PROFILE"] = "dsoc-dev"

    os.chdir(output_dir)
    cmd_line_args = ["productfetcher", "--location-file", str(json_file)]
    execute_command(cmd_line_args)

    dl_files = [file for file in Path.cwd().rglob("*") if file.name.endswith(".xml")]
    assert len(dl_files) > 0
    sdm = dl_files[0]
    assert sdm.stat().st_size == size


def test_fetch_vla_eb_from_locator(tmpdir, resource_path_root):
    """
    Pretend to fetch all the SDMs and BDFs in a VLA execution block

    :param tmpdir:
    :return:
    """

    locations_file = resource_path_root / "location_files" / "13B-014.json"
    locations_rpt = FileLocator(locations_file).locate()

    cmd_line_args = ["productfetcher", "--product-locator", "im_a_little_teapot"]
    output_dir = Path(tmpdir)
    os.chdir(output_dir)
    with patch("subprocess.Popen") as mock_proc:
        with patch(
            "productfetcher.fetchers.FileFetcher.do_fetch",
            return_value=mock_multiple_fetch(tmpdir, locations_rpt),
        ) as mock_fetch:
            execute_command(cmd_line_args)
            mock_proc.assert_called_once()
            mock_fetch.assert_not_called()

        dl_files = [file for file in output_dir.rglob("*") if file.is_file()]
        assert len(dl_files) == 33
        expected_size = 397592664
        total_size = 0

        sdm_count = 0
        bdf_count = 0
        for file in dl_files:
            with open(file, "r") as infile:
                line = infile.readline()
                total_size += int(line)

            if file.parent.name == ASDM_BIN_SUBDIR:
                bdf_count += 1
            else:
                sdm_count += 1

        assert total_size == expected_size
        assert sdm_count == 31
        assert bdf_count == 2


@pytest.mark.skipif(not REALLY_FETCH, reason="actually fetches")
def test_fetch_vla_eb_from_locator_4_realz(tmpdir):
    """
    Really fetch an execution block using a real product locator

    :return:
    """
    product_locator = "uid://evla/execblock/27561b56-4c6a-4614-bc26-67e436b5e92c"
    cmd_line_args = ["productfetcher", "--product-locator", product_locator]
    os.chdir(Path(tmpdir))
    execute_command(cmd_line_args)
    dl_files = [file for file in Path(tmpdir).rglob("*") if file.is_file()]
    assert len(dl_files) == 33
    expected_size = 397592664
    total_size = 0
    for file in dl_files:
        total_size += file.stat().st_size
    assert total_size == expected_size
    sdm_count = 0
    bdf_count = 0
    for file in dl_files:
        if file.parent.name == ASDM_BIN_SUBDIR:
            bdf_count += 1
        else:
            sdm_count += 1
    assert sdm_count == 31
    assert bdf_count == 2


# -------------------------------------------------------------------------
#
#        U T I L I T I E S
#
# -------------------------------------------------------------------------


def mock_single_fetch(tmpdir, filename: str, size: int):
    """
    Pretend to fetch a single file.

    :param tmpdir:
    :param filename:
    :param size:
    :return:
    """
    os.chdir(Path(tmpdir))
    fake_data = ""
    for _ in range(size):
        fake_data += "0"
    outfile = Path(tmpdir) / filename
    with open(outfile, "w") as out:
        out.write(fake_data)
    return outfile


def mock_multiple_fetch(tmpdir, locations_report: LocationReport) -> List[Path]:
    """
    Pretend to fetch multiple files.

    :param tmpdir:
    :param locations_report:
    :return:
    """
    output_dir = Path(tmpdir)
    os.chdir(output_dir)
    asdm_bin_subdir = output_dir / ASDM_BIN_SUBDIR
    asdm_bin_subdir.mkdir()

    disk_files = []
    for ngas_file in locations_report.files:
        size = ngas_file.size
        if ASDM_BIN_SUBDIR in ngas_file.relative_path:
            # bdf
            disk_file = asdm_bin_subdir / Path(ngas_file.relative_path).name
        else:
            # sdm
            disk_file = output_dir / ngas_file.relative_path
        with open(disk_file, "w") as out:
            out.write(f"{size}")
        disk_files.append(disk_file)

    return disk_files


def execute_command(cmd_line_args: List[str]):
    """
    Launch subprocess with these arguments.

    :param cmd_line_args:
    :return:
    """
    with subprocess.Popen(cmd_line_args):
        # no-op
        pass
